import os
import sys
import argparse

import pybedtools

import tempfile
import uuid

import numpy as np
import pandas as pd

from .util import atoi, natural_keys, bed_file

def get_df(path, category_field_num):
	
	if os.path.isdir(path):
	
		filenames = os.listdir(path)
		filenames = list(filter(bed_file, filenames))
		filenames.sort(key=natural_keys)
		
		filepaths = [os.path.join(path, filename) for filename in filenames]
		
		dfs = []
		
		for i, filepath in enumerate(filepaths):
			
			df = pd.read_csv(filepath, sep='\t', header=None, usecols=[0,1,2])
			df[3] = filenames[i].replace('.bed', '')
			
			# ensure consistent naming convention
			df[0] = [f'chr{i}' if 'chr' not in i else i for i in list(df[0])]
			
			dfs.append(df)
			
		df_merged = pd.concat(dfs)
		
		return df_merged, 3
		
	else:
		
		if category_field_num is not None:
		
			df = pd.read_csv(path, sep='\t', header=None)
			
			return df, int(category_field_num)
			
		else:
			
			print(f'Please provide a field number for identifying categories in {path}')
			sys.exit()

def odds_ratio_intersection(path_1, path_2, category_field_num_1, category_field_num_2):
	
	temp_dir = tempfile.gettempdir()
	
	df_1, category_field_num_1 = get_df(path_1, category_field_num_1)
	df_2, category_field_num_2 = get_df(path_2, category_field_num_2)
	
	categories_1 = sorted(list(set(df_1[category_field_num_1].tolist())), key=natural_keys)
	categories_2 = sorted(list(set(df_2[category_field_num_2].tolist())), key=natural_keys)
	
	path_tmp_1 = os.path.join(temp_dir, f'{str(uuid.uuid4())}.bed')
	path_tmp_2 = os.path.join(temp_dir, f'{str(uuid.uuid4())}.bed')
	
	df_1.to_csv(path_tmp_1, sep='\t', header=False, index=False)
	df_2.to_csv(path_tmp_2, sep='\t', header=False, index=False)
	
	bedtool_1 = pybedtools.BedTool(path_tmp_1)
	bedtool_2 = pybedtools.BedTool(path_tmp_2)
	
	num_fields_1 = bedtool_1.field_count()
	num_fields_2 = bedtool_2.field_count()
	
	intersection = bedtool_1.intersect(bedtool_2, wo=True)
	
	path_tmp_3 = os.path.join(temp_dir, f'{str(uuid.uuid4())}.bed')
	intersection.saveas(path_tmp_3)
	
	df = pd.read_csv(path_tmp_3, sep='\t', header=None)
	
	odds_records = []
	
	for category_1 in categories_1:
		
		for category_2 in categories_2:
		
			df_filtered = df.loc[(df[category_field_num_1] == category_1) & (df[num_fields_1 + category_field_num_2] == category_2)]
			
			# Number of overlaps and lengths
			num_overlaps = len(df_filtered.index)
			total_overlap_length = np.sum(df_filtered[len(df_filtered.columns) - 1])
			
			# Normalization metrics
			sum_length_category_1 = np.sum(df_filtered[2] - df_filtered[1])
			sum_length_category_2 = np.sum(df_filtered[num_fields_1 + 2] - df_filtered[num_fields_1 + 1])
			
			odds_records.append({'category_1': category_1, 'category_2': category_2, 'num_overlaps': num_overlaps, 'bp_overlap': total_overlap_length, 'sum_length_category_1': sum_length_category_1, 'sum_length_category_2': sum_length_category_2})
	
	df_odds = pd.DataFrame.from_records(odds_records)
	
	os.remove(path_tmp_1)
	os.remove(path_tmp_2)
	os.remove(path_tmp_3)
	
	return df_odds

# DESCRIPTION
# --------
# Script will intersect path_1 with path_2
# --------

# TF Binding BASiCS Mean
#odds_ratio_intersection -path_1 /home/data/Shared/shared_datasets/chip_seq/data/GSE61475/data/ecto/bed/input -path_2 /home/data/Shared/shared_datasets/sc_rna_seq/nlaszik/tko_mean_basics_promoter_type -path_out_csv /home/data/Shared/shared_datasets/sc_rna_seq/nlaszik/tko_mean_basics_promoter_type/tf_binding_ecto/tf.test.lod_odds.csv

def parse_arguments():
	"""
	Parses command-line arguments.
	"""
	parser = argparse.ArgumentParser()
	
	parser.add_argument("-path_1", required = True)
	parser.add_argument("-path_2", required = True)
	parser.add_argument("-path_out_csv", required = True)
	parser.add_argument("-category_field_num_1", required = False)
	parser.add_argument("-category_field_num_2", required = False)
	
	# get arguments
	return parser.parse_args()

def main():
	
	"""
	Main function for command-line usage.
	"""
	args = parse_arguments()
	
	path_1 = args.path_1
	path_2 = args.path_2
	path_out_csv = args.path_out_csv
	category_field_num_1 = args.category_field_num_1
	category_field_num_2 = args.category_field_num_2
	
	df = odds_ratio_intersection(path_1, path_2, category_field_num_1, category_field_num_2)
	
	df.to_csv(path_out_csv)
	
	pybedtools.helpers.cleanup()

if __name__ == "__main__":
	main()






import os
import sys
import re

import pandas as pd

def string_to_dict(string, item_delimiter=',', pair_delimiter=None):
	if pair_delimiter is not None:
		string_dict = {item.split(pair_delimiter)[0]:item.split(pair_delimiter)[1] for item in string.split(item_delimiter) if len(item.split(pair_delimiter)) > 1}
	else:
		string_dict = {item.split()[0]:item.split()[1] for item in string.split(item_delimiter) if len(item.split()) > 1}
		
	return string_dict

def savefile_with_readme(args, stack, path_file, data_file_string):
	
	with open(path_file, 'w') as file:
		file.write(data_file_string)
	
	dirname = os.path.dirname(path_file)
	path_readme = os.path.join(dirname, 'README.txt')
	
	path_script = stack[0][1]
	
	text_stack = ['line: ' + str(trace[2]) + ', ' + str(trace[4][0]) for i, trace in enumerate(stack) if i > 0]
	text_stack.reverse()
	
	text_readme = '\n'.join(['Datafiles generated with: ' + path_script, 'Input arguments: ' + str(args), 'Stack:\n']) + ''.join(text_stack)
	
	with open(path_readme, 'w') as file:
		file.write(text_readme)

def atoi(text):
	return int(text) if text.isdigit() else text

def natural_keys(text):
	'''
	alist.sort(key=natural_keys) sorts in human order
	http://nedbatchelder.com/blog/200712/human_sorting.html
	(See Toothy's implementation in the comments)
	'''
	return [ atoi(c) for c in re.split(r'(\d+)', text) ]

def bam_file(filename):
	ext = [".bam", ".sam"]
	return filename.endswith(tuple(ext))

def bed_file(filename):
	ext = [".bed"]
	return filename.endswith(tuple(ext))
	
def output_memmap_file(filename):
	ext = [".output.memmap"]
	return filename.endswith(tuple(ext))

def bigwig_file(filename):
	ext = [".bigWig", ".bigwig", ".bw"]
	return filename.endswith(tuple(ext))
	
def scale_num(new_max, new_min, old_max, old_min, val):
	return (new_max - new_min) * (val - old_min) / (old_max - old_min) + new_min
			
def binary_boundary_search(region_boundaries, low, high, insert_midpoint):

	if high > low:

		mid = (high + low) // 2

		if insert_midpoint >= region_boundaries[mid][0] and insert_midpoint <= region_boundaries[mid][1]:
			return mid
		elif insert_midpoint > region_boundaries[mid][1]:
			return binary_boundary_search(region_boundaries, mid + 1, high, insert_midpoint)
		else:
			return binary_boundary_search(region_boundaries, low, mid - 1, insert_midpoint)

	else:

		return -1

def create_dict_entry(input_dict, keys, entry, overwrite=True):
	
	# need an example here lol
	
	key_to_check = keys.pop(0)
	
	if len(keys) == 0:
		if key_to_check not in input_dict or overwrite:
			input_dict[key_to_check] = entry
	else:
		if key_to_check not in input_dict:
			input_dict[key_to_check] = {}
		input_dict[key_to_check] = create_dict_entry(input_dict[key_to_check], keys, entry, overwrite=overwrite)
	
	return input_dict



import os
import sys
import re

import pandas as pd

import mygene

from .util import *

def query_enhancer_atlas(path_enhancer_file):
	
	"""
	
	Queries an enhancer atlas txt file and returns a dictionary of coordinates. Can additionally pass a list of gene/transcript names/ids to filter.
	Return format: {gene_name: [[chromosome, start, stop], [chromosome, start, stop] ... ] ... }

	:param path_enhancer_file:	Path of the enhancer atlas text file
	
	Example:
	
	>>> return_dict = query_enhancer_atlas('/home/data/Shared/genome_reference_files/hg19/regions/hg19_H1_EnhancerGenePair_EnhancerAtlas2.0.txt')
	
	"""
	
	enhancer_df_ga = pd.read_csv(path_enhancer_file, sep='\t', names = ['data', 'wtf'])
	enhancer_ga_records = enhancer_df_ga.to_records()
	enhancer_gene_pairs = [enhancer['data'].replace(':','\t').replace('-','\t').replace('_','\t').replace('$','\t').split('\t') for enhancer in enhancer_ga_records]
	enhancer_ga_records = [{'chrom': enhancer_gene_pair[0], 'start': int(enhancer_gene_pair[1]), 'stop': int(enhancer_gene_pair[2]), 'genecard_id': enhancer_gene_pair[4]} for enhancer_gene_pair in enhancer_gene_pairs]
	
	enhancer_coord_dict = {}
	
	for enhancer in enhancer_ga_records:
	
		if enhancer['genecard_id'] not in enhancer_coord_dict:
			enhancer_coord_dict[enhancer['genecard_id']] = []
		
		enhancer_coord_dict[enhancer['genecard_id']].append([enhancer['chrom'], enhancer['start'], enhancer['stop']])
		
	return enhancer_coord_dict

def prepare_chromhmm(path_chrom_hmm):
	
	df = pd.read_csv(path_chrom_hmm, names=['chrom', 'start', 'stop', 'class', 'null1', 'null2', 'null3', 'null4', 'null5'], sep='\t')
	
	df = df[['chrom', 'start', 'stop', 'class']]
	
	df = df.dropna()
	
	df['start'] = df['start'].astype(int)
	df['stop'] = df['stop'].astype(int)
	
	chromhmm_list = df.to_dict('records')
	
	chromhmm_by_chrom = {}
	for i, entry in enumerate(chromhmm_list):
		
		chrom = entry['chrom']
		entry['index'] = i
		entry['midpoint'] = (entry['start'] + entry['stop'])/2
		
		if chrom not in chromhmm_by_chrom:
			chromhmm_by_chrom[chrom] = []
			
		chromhmm_by_chrom[chrom].append(entry)
		
	return chromhmm_by_chrom, chromhmm_list

def find_chromhmm_mark(chromhmm_list, i, direction, match_list=None, no_match_list=None):
	
	if no_match_list is not None and match_list is not None:
		print('please select only non-matching regions, or matching regions, cant do both')
		return False
		
	if no_match_list is None and match_list is None:
		print('please select at least one of non-matching regions or matching regions')
		return False
	
	while True:
		if direction == '+':
			if i >= len(chromhmm_list):
				return i - 1
			entry = chromhmm_list[i]
			if (no_match_list is not None and entry['class'] not in no_match_list) or (match_list is not None and entry['class'] in match_list):
				return i
			i += 1
		else:
			if i == 0:
				return i
			entry = chromhmm_list[i]
			if (no_match_list is not None and entry['class'] not in no_match_list) or (match_list is not None and entry['class'] in match_list):
				return i
			i -= 1

def find_most_likely_gene_coordinates(path_gtf, path_chromhmm, sequence_type, key, chromhmm_tss_identifier, return_type='list'):
	
	"""
	
	Queries a gtf file and returns a dictionary of coordinates. Can additionally pass a list of gene/transcript names/ids to filter.
	Return format: {key: [chromosome, gene_start, gene_stop, strand] ... }

	:param path_gtf:		Path to a gtf file

	:param sequence_type:	Which relevant genomic type is to be retrieved from the gtf file. 
							Possible values: transcript, exon, 5UTR, CDS, start_codon, stop_codon, 3UTR
	
	:param key: 			Which information to use as the key for the return dictionary.
							Possible values: gene_id, transcript_id, exon_number, exon_id, gene_name
							
	:param gene_name_filter_list:		Optional, function returns a truncated dictionary containing information for matching keys.
	
	
	Example:
	
	>>> query_gtf_coordinates('/home/data/Shared/genome_reference_files/hg19/hg19.refGene.gtf', 'transcript', 'gene_name', gene_name_filter_list = ['GAPDH'])
	
	"""
	
	#'/home/data/Shared/genome_reference_files/hg19/hg19.refGene.gtf'
	# sequence_type='transcript', key='gene_name'
	
	chromhmm_by_chrom, chromhmm_list = prepare_chromhmm(path_chromhmm)
	
	gtf_df = pd.read_csv(path_gtf, sep='\t', names=['chrom', 'gene_id_source', 'sequence_type', 'start', 'stop', 'null1', 'strand', 'null2', 'id_data'])
	gtf_df_transcripts = gtf_df.loc[gtf_df['sequence_type'] == sequence_type]
	gtf_df_transcripts_records = gtf_df_transcripts.to_records()
	
	gene_coordinate_dict = {}
	for record in gtf_df_transcripts_records:
		id_data = string_to_dict(record['id_data'], item_delimiter=';', pair_delimiter=None)
		if key in id_data:
			gene_name = id_data[key].replace('"', '')
			if gene_name not in gene_coordinate_dict:
				gene_coordinate_dict[gene_name] = []
			gene_coordinate_dict[gene_name].append({'chrom': record['chrom'], 'start': record['start'], 'stop': record['stop'], 'strand': record['strand']})
	
	real_gene_coordinate_dict = {}
	
	for gene_name in gene_coordinate_dict:
		if len(gene_coordinate_dict[gene_name]) > 0:
		
			distances_to_closest_tss = []
			in_chrom = True
			for possible_gene_coordinate in gene_coordinate_dict[gene_name]:
				
				chrom = possible_gene_coordinate['chrom']
				strand = possible_gene_coordinate['strand']
				
				if chrom not in chromhmm_by_chrom:
					in_chrom = False
					break
				
				if strand == '+':
					gene_tss = int(possible_gene_coordinate['start'])
				else:
					gene_tss = int(possible_gene_coordinate['stop'])
				
				# for alternate entry, see if TSS overlaps directly
				for chromhmm_entry in chromhmm_by_chrom[chrom]:
					i = chromhmm_entry['index']
					if gene_tss > chromhmm_entry['start'] and gene_tss < chromhmm_entry['stop']:
						break
						
				if chromhmm_list[i]['class'] == chromhmm_tss_identifier:
					# direct TSS overlap
					distances_to_closest_tss.append(i)
				else:
					
					i_nearest_tss_right = find_chromhmm_mark(chromhmm_list, i, '+', match_list=[chromhmm_tss_identifier])
					i_nearest_tss_left = find_chromhmm_mark(chromhmm_list, i, '-', match_list=[chromhmm_tss_identifier])
					
					closest_midpoint = min(chromhmm_list[i_nearest_tss_right]['midpoint'], chromhmm_list[i_nearest_tss_left]['midpoint'])
					
					distances_to_closest_tss.append(closest_midpoint)
				
			if in_chrom:
				
				most_likely_tss_i = distances_to_closest_tss.index(min(distances_to_closest_tss))
				real_gene_coordinate_dict[gene_name] = gene_coordinate_dict[gene_name][most_likely_tss_i]
				
		else:
			
			real_gene_coordinate_dict[gene_name] = gene_coordinate_dict[gene_name][0]
			
	if return_type == 'list':
		return_dict_list = {}
		for gene in real_gene_coordinate_dict:
			return_dict_list[gene] = [real_gene_coordinate_dict[gene]['chrom'], int(real_gene_coordinate_dict[gene]['start']), int(real_gene_coordinate_dict[gene]['stop']), real_gene_coordinate_dict[gene]['strand']]
		return return_dict_list
	else:
		return real_gene_coordinate_dict

def query_gtf_coordinates(path_gtf, sequence_type, key, gene_name_filter_list = None):
	
	"""
	
	Queries a gtf file and returns a dictionary of coordinates. Can additionally pass a list of gene/transcript names/ids to filter.
	Return format: {key: [chromosome, gene_start, gene_stop, strand] ... }

	:param path_gtf:		Path to a gtf file

	:param sequence_type:	Which relevant genomic type is to be retrieved from the gtf file. 
							Possible values: transcript, exon, 5UTR, CDS, start_codon, stop_codon, 3UTR
	
	:param key: 			Which information to use as the key for the return dictionary.
							Possible values: gene_id, transcript_id, exon_number, exon_id, gene_name
							
	:param gene_name_filter_list:		Optional, function returns a truncated dictionary containing information for matching keys.
	
	
	Example:
	
	>>> query_gtf_coordinates('/home/data/Shared/genome_reference_files/hg19/hg19.refGene.gtf', 'transcript', 'gene_name', gene_name_filter_list = ['GAPDH'])
	
	"""
	
	#'/home/data/Shared/genome_reference_files/hg19/hg19.refGene.gtf'
	# sequence_type='transcript', key='gene_name'
	
	gtf_df = pd.read_csv(path_gtf, sep='\t', names=['chromosome', 'gene_id_source', 'sequence_type', 'start', 'stop', 'null1', 'strand', 'null2', 'id_data'])
	
	gtf_df_transcripts = gtf_df.loc[gtf_df['sequence_type'] == sequence_type]
	gtf_df_transcripts_records = gtf_df_transcripts.to_records()
	
	gene_coordinate_dict = {string_to_dict(record['id_data'], item_delimiter=';', pair_delimiter=None)[key].replace('"', ''): [str(record['chromosome']), int(record['start']), int(record['stop']), str(record['strand'])] for record in gtf_df_transcripts_records if key in string_to_dict(record['id_data'], item_delimiter=';', pair_delimiter=None)}
	
	if gene_name_filter_list is not None:
	
		filtered_gene_coordinate_dict = {gene_name: gene_coordinate_dict[gene_name] for gene_name in gene_name_filter_list if gene_name in gene_coordinate_dict}
		print(len(gene_name_filter_list) - len(filtered_gene_coordinate_dict), 'genes not found in gtf')
		return filtered_gene_coordinate_dict
		
	else:
		
		return gene_coordinate_dict


def query_mygene_coordinates(gene_id_list, *ignore, scopes=None, genome=None):
	
	"""
	
	Queries the mygene.info database with a list of gene names/ids and returns a dictionary with their coordinates.

	:param gene_id_list:	A list of query terms, should be gene ids/names in string format
	
	:param scopes: 		A list of of types of gene ids/names that should be matched to, in descending order of priority.
				Should contain any of the following: ['ensembl', 'entrezgene', 'refseq', 'symbol', 'alias', 'name', 'other_names']
				Pass multiple in the event that your list of gene ids/names aren't all the same type.
				For example, if list contains some outdated gene symbols, add 'alias' as an additional scope.
				In the event of a duplicate, the highest priority scope will match the query term.
							
	:param genome:		'hg19' or 'hg38'
	
	
	Example:
	
	>>> query_coordinates(['YATS2'], scopes=['symbol','alias'], genome='hg19')
	
	"""
	
	# some incorrect arugments
	if ignore:  # If ignore is not empty
		raise TypeError("Incorrect arguments provided. Example usage: query_coordinates([['YEATS2']], scopes=['symbol','alias'], genome=hg19)")
	
	# check scopes
	if type(scopes) != list:
		raise TypeError("Invalid 'scopes' argument. Scopes should be a list of potential name/id types in string form that you would like to match to.\nAny of the following: ['ensembl', 'entrezgene', 'refseq', 'symbol', 'alias', 'name', 'other_names']. Pass in order of descending importance in case of duplicate return values.")
		
	valid_scopes = ['ensembl', 'entrezgene', 'refseq', 'symbol', 'alias', 'name', 'other_names']
	
	if any(scope not in valid_scopes for scope in scopes):
		raise NameError("Invalid entry in 'scopes' argument.\nShould contain any of the following: ['ensembl', 'entrezgene', 'refseq', 'symbol', 'alias', 'name', 'other_names']. Pass in order of descending importance in case of duplicate return values.")
	
	# check genome
	if genome != 'hg19' and genome != 'hg38':
		raise NameError("Invalid 'genome' argument. Should be either 'hg19' or 'hg38' as a string")

	mg = mygene.MyGeneInfo()
	
	if genome == 'hg19':
		position_field = 'genomic_pos_hg19'
	elif genome == 'hg38':
		position_field = 'genomic_pos'
		
	mg_response = mg.querymany(gene_id_list, scopes=scopes, species='human', fields=position_field)
	
	mg_response_dict = {}
	for gene in mg_response:
		if position_field in gene:
			if gene['query'] in mg_response_dict:
				# this is the duplicate entry case
				# remap this if the current gene symbol matches the current gene query
				
				if gene['query'] not in mg_response_dict:
					mg_response_dict[gene['query']] = gene[position_field]
				else:
					# check scope to see if existing entry should be rewritten
					for scope in reversed(scopes):
						if scope in gene:
							if gene[scope] == gene['query']:
								mg_response_dict[gene['query']] = gene[position_field]
			else:
				mg_response_dict[gene['query']] = gene[position_field]
		elif gene['query'] not in mg_response_dict:
			mg_response_dict[gene['query']] = 'not found'
	
	return mg_response_dict


import os
import sys
import re

import inspect
import seaborn as sns
import pandas as pd
import numpy as np
import itertools
from statannot import add_stat_annotation

import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from matplotlib import transforms
from matplotlib.artist import Artist

import matplotlib.patheffects as PathEffects

from .util import *

def savefig_with_readme(args, stack, path_fig):

	plt.savefig(path_fig)
	
	dirname = os.path.dirname(path_fig)
	path_readme = os.path.join(dirname, 'README.txt')
	
	path_script = stack[0][1]
	
	text_stack = ['line: ' + str(trace[2]) + ', ' + str(trace[4][0]) for i, trace in enumerate(stack) if i > 0]
	text_stack.reverse()
	
	text_readme = '\n'.join(['Figures generated with: ' + path_script, 'Input arguments: ' + str(args), 'Stack:\n']) + ''.join(text_stack)
	
	with open(path_readme, 'w') as file:
		file.write(text_readme)

def newline(p1, p2):
	ax = plt.gca()
	xmin, xmax = ax.get_xbound()

	if(p2[0] == p1[0]):
		xmin = xmax = p1[0]
		ymin, ymax = ax.get_ybound()
	else:
		ymax = p1[1]+(p2[1]-p1[1])/(p2[0]-p1[0])*(xmax-p1[0])
		ymin = p1[1]+(p2[1]-p1[1])/(p2[0]-p1[0])*(xmin-p1[0])

	l = mlines.Line2D([xmin,xmax], [ymin,ymax])
	ax.add_line(l)
	return l

def newtext(s, corner, **kwargs):
	ax = plt.gca()
	xmin, xmax = ax.get_xbound()
	ymin, ymax = ax.get_ybound()

	if corner == 'top-right':
		x = xmax * 0.98
		y = ymax * 0.98
		plt.text(x,y,s, horizontalalignment='right', verticalalignment = 'top', **kwargs)
	elif corner == 'top-left':
		x = xmin + (xmax - xmin)*0.02
		y = ymax * 0.98
		plt.text(x,y,s, horizontalalignment='left', verticalalignment = 'top', **kwargs)
	elif corner == 'bottom-left':
		x = xmin + (xmax - xmin)*0.02
		y = ymin + (ymax - ymin)*0.02
		plt.text(x,y,s, horizontalalignment='left', verticalalignment = 'bottom', **kwargs)
	elif corner == 'bottom-right':
		x = xmax * 0.98
		y = ymin + (ymax - ymin)*0.02
		plt.text(x,y,s, horizontalalignment='right', verticalalignment = 'bottom', **kwargs)

def add_metric_to_box_plot(df, x, y, order, ax, metric, **kwargs):
	
	unique_xs = order
	metrics = {}
	medians = {}
	pos = {}
	
	for i, unique_x in enumerate(unique_xs):
		df_x = df.loc[df[x] == unique_x]
		
		if metric == 'mean':
			metric_y = np.mean(df_x[y].tolist())
		elif metric == 'median':
			metric_y = np.median(df_x[y].tolist())
			
		median_y = np.median(df_x[y].tolist())
		
		metrics[unique_x] = metric + "=" + str(round(metric_y,3))
		medians[unique_x] = median_y
		pos[unique_x] = i
		
	# horizontalalignment='center', fontsize='x-small', color='w'
	 
	# Add it to the plot
	for tick, label in zip(pos, ax.get_xticklabels()):
		txt = ax.text(pos[tick], medians[tick] + 0.03, metrics[tick], **kwargs)
		txt.set_path_effects([PathEffects.withStroke(linewidth=2, foreground='w')])
		
def add_n_to_box_plot(df, x, y, order, ax, **kwargs):
	
	unique_xs = order
	medians = {}
	nobs = {}
	pos = {}
	
	for i, unique_x in enumerate(unique_xs):
		df_x = df.loc[df[x] == unique_x]
		nob = str(len(df_x.index))
		median_y = np.median(df_x[y].tolist())
		
		medians[unique_x] = median_y
		nobs[unique_x] = "n=" + nob
		pos[unique_x] = i
		
	# horizontalalignment='center', fontsize='x-small', color='w'
	 
	# Add it to the plot
	for tick, label in zip(pos, ax.get_xticklabels()):
		ax.text(pos[tick], medians[tick] + 0.03, nobs[tick], **kwargs)
		
def set_size(w,h, ax=None):
	""" w, h: width, height in inches """
	if not ax: ax=plt.gca()
	l = ax.figure.subplotpars.left
	r = ax.figure.subplotpars.right
	t = ax.figure.subplotpars.top
	b = ax.figure.subplotpars.bottom
	figw = float(w)/(r-l)
	figh = float(h)/(t-b)
	ax.figure.set_size_inches(figw, figh)

def plot_kde(df, x_name, filename, path_plot_dir, font_sizes=None, xlim=None, title=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	sns.kdeplot(data=df, x=x_name, ax=ax, **kwargs)
		
	if xlim is not None:
		ax.set_xlim(xlim)
	
	# plot for viewing
	path_png = os.path.join(path_plot_dir, filename + '.png')
	plt.title(title)
	plt.show()
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()	

def plot_scatter(df, x, y, filename, path_plot_dir, font_sizes=None, ylim=None, xlim=None, xlabel=None, ylabel=None, title=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	ax = sns.scatterplot(data=df, x=x, y=y, **kwargs)
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	if ylim is not None:
		ax.set_ylim(ylim)
		
	if xlim is not None:
		ax.set_xlim(xlim)	
	
	if title is not None:
		plt.title(title)
		
	plt.tight_layout()
	
	path_png = os.path.join(path_plot_dir, filename + '.png')
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	
	path_svg = os.path.join(path_plot_dir, filename + '.svg')
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	
	plt.close()

def plot_2d_kde(x, y, filename, path_plot_dir, font_sizes=None, ylim=None, xlim=None, xlabel=None, ylabel=None, title=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	ax = sns.kdeplot(x=x, y=y, **kwargs)
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	if ylim is not None:
		ax.set_ylim(ylim)
		
	if xlim is not None:
		ax.set_xlim(xlim)	
	
	if title is not None:
		plt.title(title)
		
	plt.tight_layout()
		
	path_png = os.path.join(path_plot_dir, filename + '.png')
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	
	path_svg = os.path.join(path_plot_dir, filename + '.svg')
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	
	plt.close()


def plot_scatter_list(x, y, filename, path_plot_dir, font_sizes=None, ylim=None, xlim=None, xlabel=None, ylabel=None, title=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	ax = sns.scatterplot(x=x, y=y, **kwargs)
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	if ylim is not None:
		ax.set_ylim(ylim)
		
	if xlim is not None:
		ax.set_xlim(xlim)	
	
	if title is not None:
		plt.title(title)
		
	plt.show()
	path_png = os.path.join(path_plot_dir, filename + '.png')
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()
	
def plot_hist(df, x_name, filename, path_plot_dir, font_sizes=None, ylim=None, xlim=None, title=None, xlabel=None, ylabel=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	sns.histplot(data=df, x=x_name, ax=ax, **kwargs)
	
	if ylim is not None:
		ax.set_ylim(ylim)
		
	if xlim is not None:
		ax.set_xlim(xlim)
		
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	# plot for viewing
	path_png = os.path.join(path_plot_dir, filename + '.png')
	plt.title(title)
	plt.show()
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()	

def plot_hist_list(data_list, filename, path_plot_dir, font_sizes=None, ylim=None, xlim=None, title=None, xlabel=None, ylabel=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	sns.histplot(data_list, ax=ax, **kwargs)
	
	if ylim is not None:
		ax.set_ylim(ylim)
		
	if xlim is not None:
		ax.set_xlim(xlim)
		
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	# plot for viewing
	path_png = os.path.join(path_plot_dir, filename + '.png')
	plt.title(title)
	plt.show()
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()
	
def plot_violin(df, x, y, order, filename, path_plot_dir, font_sizes=None, fig_width=None, fig_height=None, palette_name=None, metric=None, stat_test=None, rotate_x=False, xlabel=None, ylabel=None, ylim=None, hue=None, hue_order=None, pad_top=False, pad_bottom=False, pad_left=False, pad_right=False, title=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	if pad_top and pad_bottom:
		fig, ax = plt.subplots(figsize = (8, 10))
	elif (pad_top and not pad_bottom) or (pad_bottom and not pad_top):
		fig, ax = plt.subplots(figsize = (8, 8))
	else:
		fig, ax = plt.subplots()
		
	if pad_top:
		fig.subplots_adjust(top=0.55)
	if pad_bottom:
		fig.subplots_adjust(bottom=0.45)
	if pad_left:
		fig.subplots_adjust(left=0.25)
	if pad_right:
		fig.subplots_adjust(right=0.85)
	
	#palette=sns.color_palette(palette_name, n_colors=len(order))
		
	# in a violin plot, 'x' serves as the hue
		
	if hue is not None and hue_order is not None:
		# for when there are inner categories
		sns.violinplot(data=df, x=x, y=y, ax=ax, order=order, hue=hue, hue_order=hue_order, **kwargs)
		if stat_test is not None:
			inner_categories = list(set(df[hue].tolist()))
			x_pairs = [((outer_category, inner_pair[0]), (outer_category, inner_pair[1])) for inner_pair in list(itertools.combinations(inner_categories, 2)) for outer_category in order]
			add_stat_annotation(ax, data=df, x=x, y=y, order=order, box_pairs=x_pairs, hue=hue, test=stat_test, text_format='star', fontsize=SMALL_SIZE, loc='inside', verbose=0, linewidth=1, color='black')
	else:
		# only one set of categories
		sns.violinplot(data=df, x=x, y=y, ax=ax, order=order, **kwargs)
		if stat_test is not None:
			x_pairs = list(itertools.combinations(order, 2))
			add_stat_annotation(ax, data=df, x=x, y=y, order=order, box_pairs=x_pairs, test=stat_test, text_format='star', fontsize=SMALL_SIZE, loc='inside', verbose=0, linewidth=1, color='black')
	
	if ylim is not None:
		ax.set_ylim(ylim)
	
	if metric is not None:
		add_metric_to_box_plot(df, x, y, order, ax, metric)
	
	path_png = os.path.join(path_plot_dir, filename + '.png')
	path_svg = path_png.replace('.png', '.svg')
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	ax.spines['right'].set_visible(False)
	ax.spines['top'].set_visible(False)
	
	if rotate_x:
		plt.xticks(rotation=45, ha='right')
	
	if title is not None:
		plt.title(title)
		
	plt.tight_layout()
		
	plt.show()
	
	if fig_width is not None and fig_height is not None:
		set_size(fig_width, fig_height, ax)
	
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()	
	
def plot_box(df, x, y, order, filename, path_plot_dir, font_sizes=None, fig_width=None, fig_height=None, palette_name=None, metric=None, add_n=False, add_stat=False, rotate_x=False, xlabel=None, ylabel=None, ylim=None, hue=None, hue_order=None, pad_top=False, pad_bottom=False, pad_left=False, pad_right=False, title=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	if pad_top:
		fig.subplots_adjust(top=0.55)
	if pad_bottom:
		fig.subplots_adjust(bottom=0.25)
	if pad_left:
		fig.subplots_adjust(left=0.25)
	if pad_right:
		fig.subplots_adjust(right=0.85)
	
	#palette=sns.color_palette(palette_name, n_colors=len(order))
		
	if hue is not None and hue_order is not None:
		sns.boxplot(data=df, x=x, y=y, ax=ax, order=order, hue=hue, hue_order=hue_order, **kwargs)
		if add_stat:
			inner_categories = list(set(df[hue].tolist()))
			x_pairs = [((outer_category, inner_pair[0]), (outer_category, inner_pair[1])) for inner_pair in list(itertools.combinations(inner_categories, 2)) for outer_category in order]
			add_stat_annotation(ax, data=df, x=x, y=y, order=order, box_pairs=x_pairs, hue=hue, test='t-test_ind', text_format='star', fontsize=SMALL_SIZE, loc='outside', verbose=0, linewidth=1, color='black')
	else:
		sns.boxplot(data=df, x=x, y=y, ax=ax, order=order, **kwargs)
		if add_stat:
			x_pairs = list(itertools.combinations(order, 2))
			add_stat_annotation(ax, data=df, x=x, y=y, order=order, box_pairs=x_pairs, test='t-test_ind', text_format='full', fontsize=SMALL_SIZE, loc='outside', verbose=0, linewidth=1, color='black')
		
	if ylim is not None:
		ax.set_ylim(ylim)
	
	if add_n:
		add_n_to_box_plot(df, x, y, order, ax)
	
	path_png = os.path.join(path_plot_dir, filename + '.png')
	path_svg = path_png.replace('.png', '.svg')
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
	
	ax.spines['right'].set_visible(False)
	ax.spines['top'].set_visible(False)
	
	if metric is not None:
		add_metric_to_box_plot(df, x, y, order, ax, metric, ha='center')
	
	if rotate_x:
		plt.xticks(rotation=45, ha='right')
	
	if title is not None:
		plt.title(title)
	
	if fig_width is not None and fig_height is not None:
		set_size(fig_width, fig_height, ax)
		
	plt.tight_layout()
	plt.show()
	
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()
	
def scale_num(new_max, new_min, old_max, old_min, val):
	return (new_max - new_min) * (val - old_min) / (old_max - old_min) + new_min
	
def rainbow_text(x, y, strings, colors, weights, ha, spacing, **kw):
	"""
	Take a list of strings ``strings`` and colors ``colors`` and place them next to each
	other, with text strings[i] being shown in color colors[i].
	
	x and y are axes coordinates, not data coordinates
	"""
	t = plt.gca().transAxes
	fig = plt.gcf()
	plt.show()
	
	#horizontal version
	newline_offset = 0
	first_t = False
	line_widths = []
	line_width = 0
	text_objects = []
	
	if ha == 'left':
		for s,c,w in zip(strings, colors, weights):
				
			if s != '\n':
				text = plt.text(x, y, s+" ", color=c, transform=t, fontweight=w, **kw)
				text.draw(fig.canvas.get_renderer())
				
				ex = text.get_window_extent()
				t = transforms.offset_copy(text._transform, x=ex.width, units='dots')
				
				if not first_t:
					first_t = text._transform
				
			else:
				# reset the x transform, add onto y transform
				# 1.5 spacing
				newline_offset += (ex.height * spacing)
				t = transforms.offset_copy(first_t, y=-newline_offset, units='dots')
				
	elif ha == 'center':
	
		# draw entire line first to get line widths
		for s,c,w in zip(strings, colors, weights):
				
			if s != '\n':
				text = plt.text(x, y, s+" ", color=c, transform=t, fontweight=w, **kw)
				text.draw(fig.canvas.get_renderer())
				text_objects.append(text)
				
				ex = text.get_window_extent()
				t = transforms.offset_copy(text._transform, x=ex.width, units='dots')
				line_width += ex.width
				
				if not first_t:
					first_t = text._transform
				
			else:
				# reset the x transform, add onto y transform
				# 1.5 spacing
				newline_offset += (ex.height * spacing)
				t = transforms.offset_copy(first_t, y=-newline_offset, units='dots')
				line_widths.append(line_width)
				line_width = 0
				
		line_widths.append(line_width)
		line_h_offsets = [-width/2 for width in line_widths]
		
		t = plt.gca().transAxes
		newline_offset = 0
		
		# delete now that we have offset
		for text_object in text_objects:
			Artist.remove(text_object)
			
		i_line = 0
		t = transforms.offset_copy(first_t, x=line_h_offsets[i_line], units='dots')
		i_line += 1
			
		# redraw
		for s,c,w in zip(strings, colors, weights):
			
			if s != '\n':
				text = plt.text(x, y, s+" ", color=c, transform=t, fontweight=w, **kw)
				text.draw(fig.canvas.get_renderer())
				
				ex = text.get_window_extent()
				t = transforms.offset_copy(text._transform, x=ex.width, units='dots')
				line_width += ex.width
				
			else:
				# reset the x transform, add onto y transform
				# 1.5 spacing
				newline_offset += (ex.height * spacing)
				t = transforms.offset_copy(first_t, y=-newline_offset, x=line_h_offsets[i_line], units='dots')
				i_line += 1
	
def plot_line(df, x_name, y_name, filename, path_plot_dir, font_sizes=None, hue=None, custom_ci_names=None, xlim=None, ylim=None, title=None, fig_width=None, fig_height=None, xticks=None, xlabels=None, xlabel=None, yticks=None, ylabels=None, ylabel=None, legend_text=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	
	if hue is not None:
		
		hue_order = list(set(df[hue].tolist()))
		hue_order.sort(key=natural_keys)
		palette=sns.color_palette("mako", n_colors=len(hue_order))
		palette.reverse()
		
		# ci = 1.96 * np.std(y)/np.sqrt(len(x))
	
		if custom_ci_names is not None:
			lower_bound_name = custom_ci_names[0]
			upper_bound_name = custom_ci_names[1]
			for i, group in enumerate(hue_order):
				df_hue = df.loc[df[hue] == group]
				x = df_hue[x_name].tolist()
				lower_bound = df_hue[lower_bound_name].tolist()
				upper_bound = df_hue[upper_bound_name].tolist()
				plt.fill_between(x, lower_bound, upper_bound, color=palette[i], alpha=.33)
				
		
		# make a custom legend. we have the colors
		if legend_text is not None:
			sns.lineplot(data=df, x=x_name, y=y_name, hue=hue, hue_order=hue_order, ax=ax, palette=palette, legend=False, **kwargs)
			num_words = len(legend_text.split(' '))
			legend_lines = ' '.join(['|' for i in hue_order])
			legend_descriptor = r'Low $\rightarrow$ High'
			
			legend_text = f'{legend_text} \n {legend_lines} \n {legend_descriptor}'.split(' ')
			
			legend_colors = [(0.0, 0.0, 0.0)] * (num_words + 1) + palette + [(0.0, 0.0, 0.0)] * 4
			legend_weights = ['normal'] * (num_words + 1) + ['bold'] * len(palette) + ['normal'] * 4
			
			ha = 'center'
			spacing = 1.5
			
			rainbow_text(1.2, 0.5, legend_text, legend_colors, legend_weights, ha, spacing)
			
		else:
			sns.lineplot(data=df, x=x_name, y=y_name, hue=hue, hue_order=hue_order, ax=ax, palette=palette, **kwargs)
			plt.legend()
		
	else:
		sns.lineplot(data=df, x=x_name, y=y_name, ax=ax, **kwargs)
	
	if ylim is not None:
		ax.set_ylim(ylim)
	
	if xlim is not None:
		ax.set_xlim(xlim)
		
	ax.spines['right'].set_visible(False)
	ax.spines['top'].set_visible(False)
	
	if xticks is not None:
		ax.set_xticks(xticks)
		if xlabels is not None:
			ax.set_xticklabels(xlabels)	
		
	if yticks is not None:
		ax.set_yticks(yticks)
		if ylabels is not None:
			ax.set_yticklabels(ylabels)
			
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
		
	# plot for viewing
	path_png = os.path.join(path_plot_dir, filename + '.png')
	path_svg = path_png.replace('.png', '.svg')
	
	plt.title(title)
	
	# add some padding on right side for custom legend
	if legend_text is not None:
		plt.tight_layout(rect = (0, 0, 1, 1))
	
	plt.show()
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	
	# plot for final figure
	if fig_width is not None and fig_height is not None:
		set_size(fig_width, fig_height, ax)
		
		path_png = os.path.join(path_plot_dir, filename + '.for_figure.png')
		path_svg = path_png.replace('.png', '.svg')
		
		savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
		savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
		
	plt.close()
	
def plot_bar(df, x, y, order, filename, path_plot_dir, font_sizes = None, fig_width=None, fig_height=None, add_n=False, add_stat=False, rotate_x=False, xlabel=None, ylabel=None, ylim=None, hue=None, hue_order=None, **kwargs):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	fig.subplots_adjust(left=0.25, right=0.85, top=0.55, bottom=0.25)
	
	if hue is not None and hue_order is not None:
		sns.barplot(data=df, x=x, y=y, ax=ax, hue=hue, order=order, hue_order=hue_order, **kwargs)
		if add_stat:
			inner_categories = list(set(df[hue].tolist()))
			x_pairs = [((outer_category, inner_pair[0]), (outer_category, inner_pair[1])) for inner_pair in list(itertools.combinations(inner_categories, 2)) for outer_category in order]
			add_stat_annotation(ax, data=df, x=x, y=y, order=order, box_pairs=x_pairs, hue=hue, test='t-test_ind', text_format='star', fontsize=SMALL_SIZE, loc='outside', verbose=0, linewidth=1, color='black')
	else:
		sns.barplot(data=df, x=x, y=y, ax=ax, order=order, hue=hue, **kwargs)
		if add_stat:
			x_pairs = list(itertools.combinations(order, 2))
			add_stat_annotation(ax, data=df, x=x, y=y, order=order, box_pairs=x_pairs, test='t-test_ind', text_format='star', fontsize=SMALL_SIZE, loc='outside', verbose=0, linewidth=1, color='black')
	
	path_png = os.path.join(path_plot_dir, filename + '.png')
	path_svg = path_png.replace('.png', '.svg')
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
		
	if ylim is not None:
		ax.set_ylim(ylim)
		
	if add_n:
		add_n_to_bar_plot(df, x, y, order, ax)
	
	ax.spines['right'].set_visible(False)
	ax.spines['top'].set_visible(False)
	
	if rotate_x:
		plt.xticks(rotation=45, ha='right')
	
	plt.title(filename)
	plt.show()
	
	if fig_width is not None and fig_height is not None:
		set_size(fig_width, fig_height, ax)
	
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()
	
def add_n_to_bar_plot(df, x, y, order, ax):
	
	unique_xs = order
	means = {}
	nobs = {}
	pos = {}
	
	for i, unique_x in enumerate(unique_xs):
		df_x = df.loc[df[x] == unique_x]
		nob = str(len(df_x.index))
		mean_y = np.mean(df_x[y].tolist())
		
		means[unique_x] = mean_y
		nobs[unique_x] = "n=" + nob
		pos[unique_x] = i
	 
	# Add it to the plot
	for tick, label in zip(pos, ax.get_xticklabels()):
		ax.text(pos[tick], means[tick] + 0.1, nobs[tick], horizontalalignment='left', color='black', fontsize=SMALL_SIZE, rotation=45)

def add_stdev(df, x, y, order, ax, colors=None):
	
	lines = {}
	means = {}
	pos = {}
	colors_dict = {}
	
	for i, unique_x in enumerate(order):
	
		df_x = df.loc[df[x] == unique_x]
		
		mean_y = np.mean(df_x[y].tolist())
		#standard_err = stats.sem(df_x[y].tolist())
		standard_err = np.std(df_x[y].tolist())
		
		means[unique_x] = mean_y
		lines[unique_x] = [[i, mean_y-standard_err],[i, mean_y+standard_err]]
		pos[unique_x] = i
		colors_dict[unique_x] = colors[i]
	 
	# Add it to the plot
	for tick, label in zip(pos, ax.get_xticklabels()):
		if colors is not None:
			newline(lines[tick][0], lines[tick][1], color=colors_dict[tick])
			plt.plot(pos[tick], means[tick], marker='D', markersize=3, color=colors_dict[tick])
		else:
			newline(lines[tick][0], lines[tick][1])
			plt.plot(pos[tick], means[tick], marker='D', markersize=3)


def plot_mean_and_sem(df, x, y, order, filename, path_plot_dir, font_sizes=None, fig_width=None, fig_height=None, colors=None, rotate_x=False, xlabel=None, ylabel=None, ylim=None):
	
	if font_sizes is not None:
		if font_sizes == 'powerpoint':
			set_powerpoint_sizes()
		elif font_sizes == 'article':
			set_article_sizes()
	
	fig, ax = plt.subplots()
	fig.subplots_adjust(left=0.25, right=0.85, top=0.55, bottom=0.25)
	
	plt.xticks(ticks=list(range(len(order))), labels=order)
	
	add_stdev(df, x, y, order, ax, colors=colors)
	
	path_png = os.path.join(path_plot_dir, filename + '.png')
	path_svg = path_png.replace('.png', '.svg')
	
	if xlabel is not None:
		ax.set_xlabel(xlabel)
		
	if ylabel is not None:
		ax.set_ylabel(ylabel)
		
	if ylim is not None:
		ax.set_ylim(ylim)
		
	ax.spines['right'].set_visible(False)
	ax.spines['top'].set_visible(False)
	
	if rotate_x:
		plt.xticks(rotation=45, ha='right')
	
	plt.title(filename)
	plt.show()
	
	if fig_width is not None and fig_height is not None:
		set_size(fig_width, fig_height, ax)
	
	savefig_with_readme(str(sys.argv), inspect.stack(), path_svg)
	savefig_with_readme(str(sys.argv), inspect.stack(), path_png)
	plt.close()



import os, sys, re, argparse, random

import pandas as pd
import numpy as np
import pyranges as pr

from tqdm import tqdm

from joblib import Parallel, delayed, dump, load, parallel_backend

from collections import defaultdict

from .util import *

def get_df(path, category_field_num):

	if os.path.isdir(path):
	
		filenames = os.listdir(path)
		filenames = list(filter(bed_file, filenames))
		filenames.sort(key=natural_keys)
		
		filepaths = [os.path.join(path, filename) for filename in filenames]
		
		dfs = []
		
		for i, filepath in enumerate(filepaths):
			
			# Column names required for pyranges
			df = pd.read_csv(filepath, sep='\t', header=None, usecols=[0,1,2], names=['Chromosome', 'Start', 'End'])
			df[3] = filenames[i].replace('.bed', '')
			
			# ensure consistent naming convention
			df['Chromosome'] = [f'chr{i}' if 'chr' not in i else i for i in list(df['Chromosome'])]
			
			dfs.append(df)
			
		df_merged = pd.concat(dfs)
		
		return df_merged, 3
		
	else:
		
		if category_field_num is not None:
		
			df = pd.read_csv(path, sep='\t', header=None)
			
			# Column names required for pyranges
			df.rename(columns={0: 'Chromosome', 1:'Start', 2:'End'}, inplace=True)
			
			# ensure consistent naming convention
			df['Chromosome'] = [f'chr{i}' if 'chr' not in i else i for i in list(df['Chromosome'])]
			
			return df, int(category_field_num)
			
		else:
			
			print(f'Please provide a field number for identifying categories in {path}')
			sys.exit()

def create_slices(input_list, num_slices):
	"""
	Create slices of the input list into a specified number of slices.
	
	Args:
	input_list (list): The list to be sliced.
	num_slices (int): The number of slices to create.
	
	Returns:
	list: A list containing the slices.
	"""
	# Calculate the size of each slice
	slice_size = len(input_list) // num_slices
	remainder = len(input_list) % num_slices
	
	slices = []
	start = 0
	
	for i in range(num_slices):
		# Calculate the end index for the current slice
		end = start + slice_size + (1 if i < remainder else 0)
		# Append the current slice to the slices list
		slices.append(input_list[start:end])
		# Update the start index for the next slice
		start = end
	
	return slices

def calculate_distributions(df):
	"""
	Calculate the distributions of region lengths and distances between regions, separately for each chromosome.
	"""
	chrom_distributions = defaultdict(lambda: {'lengths': [], 'distances': []})
	
	for chrom in df['Chromosome'].unique():
		chrom_df = df[df['Chromosome'] == chrom].copy()
		chrom_df['distance'] = chrom_df['Start'].shift(-1) - chrom_df['End']
		
		chrom_distributions[chrom]['lengths'] = chrom_df['length'].values
		chrom_distributions[chrom]['distances'] = chrom_df['distance'].values
		
	return chrom_distributions

def generate_random_overlap_slice(slice, distributions_to_shuffle, df_static, chrom_sizes, mask_dict):
	"""
	Generate a random track with the given lengths and distances, avoiding mask regions.
	"""
	overlaps = []
	for i_slice in slice:
		
		random_region_records = []
		
		for chrom, size in chrom_sizes.items():
			
			lengths = distributions_to_shuffle[chrom]['lengths']
			distances = distributions_to_shuffle[chrom]['distances']
			
			random.shuffle(lengths)
			random.shuffle(distances)
			
			current_position = 0
			chrom_mask = sorted(mask_dict.get(chrom, []), key=lambda x: x[0])
			mask_index = 0
			mask_length = len(chrom_mask)
			
			for length, distance in zip(lengths, distances):
				start = current_position + distance
				end = start + length
				
				while mask_index < mask_length and chrom_mask[mask_index][1] < start:
					mask_index += 1
				
				while (mask_index < mask_length and 
					   (chrom_mask[mask_index][0] <= start <= chrom_mask[mask_index][1] or 
						chrom_mask[mask_index][0] <= end <= chrom_mask[mask_index][1])):
					start = chrom_mask[mask_index][1] + 1
					end = start + length
					mask_index += 1
				
				if end <= size:
					random_region_records.append({'Chromosome': chrom, 'Start': int(start), 'End': int(end)})
					current_position = end
				else:
					break
		
		df_shuffle = pd.DataFrame.from_records(random_region_records)
		
		gr1 = pr.PyRanges(df_shuffle)
		gr2 = pr.PyRanges(df_static)
		
		# Perform intersection
		df_intersected = gr1.join(gr2, how='left', report_overlap=True).df
		df_intersected = df_intersected[df_intersected['Overlap'] > 0]
		
		num_overlap = len(df_intersected['Overlap'])
		sum_overlap = np.sum(df_intersected['Overlap'])
		
		overlaps.append({'num_overlaps': num_overlap, 'length_overlaps': sum_overlap})
		
	return overlaps
	
def generate_random_overlaps(df_shuffle, df_static, num_permutations, mask_dict, chrom_sizes, num_processes):
		
	# get distribution of distances and lengths
	distributions_to_shuffle = calculate_distributions(df_shuffle)
	
	# this should be parallelized here
	input_list = list(range(num_permutations))  # Example list of length 1000
	slices = create_slices(input_list, num_processes)
	
	with parallel_backend("loky", inner_max_num_threads=2):
		random_overlap_list = Parallel(n_jobs=num_processes, verbose=10, pre_dispatch="all")(delayed(generate_random_overlap_slice)(slice, distributions_to_shuffle, df_static, chrom_sizes, mask_dict) for slice in slices)
	
	random_overlap_records = [item for sublist in random_overlap_list for item in sublist]
	
	df_random_overlaps = pd.DataFrame.from_records(random_overlap_records)

	return df_random_overlaps

def calculate_p(random_list, enrichment):
	
	num_bins = max(random_list) - min(random_list)
	hist, bin_edges = np.histogram(random_list, bins=num_bins, density=True)
	
	x_d = np.linspace(min(random_list), max(random_list), num_bins)
	
	p_value = sum([hist[i] for i, x in enumerate(x_d) if x > enrichment])
		
	return p_value

def enrichment(path_1, path_2, category_field_num_1, category_field_num_2, path_mask, path_sizes, num_permutations, num_processes):
	
	df_mask = pd.read_csv(path_mask, sep='\t', header=None)
	mask_dict = {}
	for record in df_mask.to_dict('records'):
		if record[0] not in mask_dict:
			mask_dict[record[0]] = []
		mask_dict[record[0]].append((record[1], record[2]))
	
	# initialize distances and chromosomes dicts
	df_chrom_sizes = pd.read_csv(path_sizes, sep='\t', header=None)
	chrom_sizes = {}
	for record in df_chrom_sizes.to_dict('records'):
		chrom_sizes[record[0]] = record[1]
	
	df_1, category_field_num_1 = get_df(path_1, category_field_num_1)
	df_2, category_field_num_2 = get_df(path_2, category_field_num_2)
	
	df_1['length'] = df_1['End'] - df_1['Start']
	df_2['length'] = df_2['End'] - df_2['Start']
	
	categories_1 = sorted(list(set(df_1[category_field_num_1].tolist())), key=natural_keys)
	categories_2 = sorted(list(set(df_2[category_field_num_2].tolist())), key=natural_keys)
	
	enrichment_records = []
	
	for category_1 in categories_1:
		
		for category_2 in categories_2:
			
			df_1_filt = df_1.loc[df_1[category_field_num_1] == category_1]
			df_2_filt = df_2.loc[df_2[category_field_num_2] == category_2]
			
			sum_length_1 = np.sum(df_1_filt['length'])
			sum_length_2 = np.sum(df_2_filt['length'])
			
			gr1 = pr.PyRanges(df_1_filt)
			gr2 = pr.PyRanges(df_2_filt)
			
			df_intersected = gr1.join(gr2, how='left', report_overlap=True).df
			df_intersected = df_intersected[df_intersected['Overlap'] > 0]
			
			num_overlaps = len(df_intersected['Overlap'])
			length_overlaps = np.sum(df_intersected['Overlap'])
			
			####################
			# RANDOM INTERSECTIONS
			####################
			shuffle = 2
			if shuffle == 1:
				df_shuffle = df_1_filt
				df_static = df_2_filt
			else:
				df_shuffle = df_2_filt
				df_static = df_1_filt
				
			print(f'Generating {num_permutations} permutations, not in masked region for {category_1} and {category_2} comparison')
			df_random_overlaps = generate_random_overlaps(df_shuffle, df_static, num_permutations, mask_dict, chrom_sizes, num_processes)
			
			mean_length_random_overlaps = np.mean(df_random_overlaps['length_overlaps'])
			mean_num_permutationsom_overlaps = np.mean(df_random_overlaps['num_overlaps'])
			
			length_enrichment = np.log2(length_overlaps / mean_length_random_overlaps)
			num_enrichment = np.log2(num_overlaps / mean_num_permutationsom_overlaps)
			
			length_p = calculate_p(df_random_overlaps['length_overlaps'], length_enrichment)
			num_p = calculate_p(df_random_overlaps['num_overlaps'], num_enrichment)
				
			enrichment_records.append({'category_1': category_1, 'category_2': category_2, 'num_overlaps_log2_enrichment': num_enrichment, 'length_overlaps_log2_enrichment': length_enrichment, 'length_p': length_p, 'num_p': num_p})
	
	df_enrichment = pd.DataFrame.from_records(enrichment_records)
	
	return df_enrichment

# DESCRIPTION
# --------
# Script will calculate enrichment of path_1 vs. path_2
# --------

# Clock CpGs and ChromHMM
# enrichment -path_1 /home/data/Shared/shared_datasets/wgbs/nlaszik/horvath_clock_skin_blood/separated -path_2 /home/data/Shared/shared_datasets/chip_seq/data/roadmap_epigenomics_chromhmm/data/E016_25_imputed12marks_dense.bed -category_field_num_2 3 -path_mask /home/data/Shared/genome_reference_files/hg19/hg19_gap.bed -path_sizes /home/data/Shared/genome_reference_files/hg19/hg19.sizes -path_out_csv /home/data/Shared/shared_datasets/wgbs/nlaszik/horvath_clock_skin_blood/test.enrichment.csv -num_processes 64 -num_permutations 1000

def parse_arguments():
	"""
	Parses command-line arguments.
	"""
	parser = argparse.ArgumentParser()
	
	parser.add_argument("-path_1", required = True)
	parser.add_argument("-path_2", required = True)
	parser.add_argument("-path_mask", required = True)
	parser.add_argument("-path_sizes", required = True)
	parser.add_argument("-num_permutations", required = True)
	parser.add_argument("-num_processes", required = True)
	parser.add_argument("-path_out_csv", required = True)
	parser.add_argument("-category_field_num_1", required = False)
	parser.add_argument("-category_field_num_2", required = False)
	
	# get arguments
	return parser.parse_args()

def main():
	
	"""
	Main function for command-line usage.
	"""
	args = parse_arguments()
	
	path_1 = args.path_1
	path_2 = args.path_2
	path_mask = args.path_mask
	path_sizes = args.path_sizes
	num_permutations = int(args.num_permutations)
	num_processes = int(args.num_processes)
	category_field_num_1 = args.category_field_num_1
	category_field_num_2 = args.category_field_num_2
	path_out_csv = args.path_out_csv
	
	df = enrichment(path_1, path_2, category_field_num_1, category_field_num_2, path_mask, path_sizes, num_permutations, num_processes)
	
	df.to_csv(path_out_csv)

if __name__ == "__main__":
	main()








# main.py
#from .gene import query_enhancer_atlas, find_most_likely_gene_coordinates, query_gtf_coordinates, query_mygene_coordinates
from .odds_ratio_intersection import odds_ratio_intersection
from .enrichment import enrichment
from .cluster_regions import cluster_regions

################
# FUNCTIONS
################

def cluster_regions(df, value_threshold_pair, distance_threshold, min_num_values, suffix=None):

	return cluster_regions(df, value_threshold_pair, distance_threshold, min_num_values, suffix)

def enrichment(path_1, path_2, category_field_num_1, category_field_num_2, path_mask, path_sizes, num_permutations, num_processes):

	return enrichment(path_1, path_2, category_field_num_1, category_field_num_2, path_mask, path_sizes, num_permutations, num_processes)

def odds_ratio_intersection(path_1, path_2, category_field_num_1, category_field_num_2):

	return odds_ratio_intersection(path_1, path_2, category_field_num_1, category_field_num_2)



import os
import sys
import pandas as pd
from tqdm import tqdm
import argparse

# cluster_regions --path_bed /home/data/Shared/shared_datasets/wgbs/nlaszik/wt_dko_wgbs_intersect/GSM1112840.HUES64.WGBS_GSM3662265.HUES64_DKO.WGBS_difference.bed --value_thresholds="-1,-0.5,0,0.5,1" --distance_threshold 1000 --path_out_dir /home/data/Shared/shared_datasets/wgbs/nlaszik/wt_dko_wgbs_intersect/dmr_test --suffix _dko --min_num_values 3


# cluster_regions --path_bed /home/data/Shared/shared_datasets/wgbs/nlaszik/wt_dko_wgbs_intersect/GSM1112840.HUES64.WGBS_GSM3662265.HUES64_DKO.WGBS_difference.bed --value_thresholds="-1,-0.3,0,0.3,1" --distance_threshold 1000 --path_out_dir /home/data/Shared/shared_datasets/wgbs/nlaszik/wt_dko_wgbs_intersect/dmr_test --suffix _tko --min_num_values 3

def cluster_regions(df, value_threshold_pair, distance_threshold, min_num_values, suffix=None):
	
	##########
	# 1. Filter differentially methylated CpGs
	# This can easily be adjusted to whatever you would like... just copy the script!
	##########
	df_filt = df.loc[(df['value'] >= value_threshold_pair[0]) & (df['value'] <= value_threshold_pair[1])]
	
	##########
	# 2. Find DMRs by clustering adjacent CpGs
	##########
	clustered_regions = get_clusters(df_filt, distance_threshold)
	
	# Convert DMRs to DataFrame
	df_clustered_regions = pd.DataFrame(clustered_regions, columns=['chromosome', 'start', 'end', 'total_value', 'num_values'])
	df_clustered_regions['category'] = f'{value_threshold_pair[0]}_{value_threshold_pair[1]}{suffix}'
	
	##########
	# 3. Filter by number of CpGs in DMR
	##########
	df_clustered_regions = df_clustered_regions.loc[df_clustered_regions['num_values'] > min_num_values]
	
	return df_clustered_regions

def get_clusters(df, distance_threshold):
	
	regions = []
	current_region = None
	previous_end = None
	previous_chrom = None
	value_counter = 0
	
	df_records = df.to_records('dict')
	
	previous_chrom = df_records[0]['chromosome']
	
	for row in tqdm(df_records):
		# Check if the current CpG is within the distance cutoff from the previous one
		
		this_chrom = row['chromosome']
		
		if current_region is None or (previous_end is not None and row['start'] - previous_end > distance_threshold) or this_chrom != previous_chrom:
			if current_region is not None:
				current_region.append(value_counter)
				regions.append(current_region)
			current_region = [row['chromosome'], row['start'], row['end'], row['value']]
			value_counter = 1
		else:
			current_region[2] = row['end']  # Extend the current DMR
			current_region[3] += row['value']  # Accumulate methylation diff
			value_counter += 1
			
		previous_end = row['end']
		previous_chrom = row['chromosome']
	
	# Add the last DMR if it exists
	if current_region is not None:
		current_region.append(value_counter)
		regions.append(current_region)
		
	return regions
	
	
def parse_arguments():
	"""
	Parses command-line arguments.
	"""
	parser = argparse.ArgumentParser()
	parser.add_argument("--path_bed", required = True)
	parser.add_argument("--value_thresholds", required = True)
	parser.add_argument("--distance_threshold", required = True)
	parser.add_argument("--path_out_dir", required = True)
	parser.add_argument("--suffix", required = False)
	parser.add_argument("--min_num_values", required = False, default=1)
	
	# get arguments
	return parser.parse_args()

def main():
	
	"""
	Main function for command-line usage.
	"""
	args = parse_arguments()
	
	path_bed = args.path_bed
	value_thresholds = args.value_thresholds
	distance_threshold = int(args.distance_threshold)
	path_out_dir = args.path_out_dir
	suffix = args.suffix
	min_num_values = int(args.min_num_values)
	
	os.makedirs(path_out_dir, exist_ok=True)
	
	value_thresholds = [float(val) for val in value_thresholds.split(',')]
	
	value_threshold_pairs = [[value_thresholds[i], value_thresholds[i+1]] for i in range(len(value_thresholds) - 1)]
	
	df = pd.read_csv(path_bed, sep='\t', header=None, names=['chromosome', 'start', 'end', 'value'], usecols=[0,1,2,3])
	
	cluster_dfs = [cluster_regions(df, value_threshold_pair, distance_threshold, min_num_values, suffix) for value_threshold_pair in value_threshold_pairs]
	
	for i, cluster_df in enumerate(cluster_dfs):
		
		path_out = os.path.join(path_out_dir, f'{value_threshold_pairs[i][0]}_{value_threshold_pairs[i][1]}{suffix}.bed')
		cluster_df.to_csv(path_out, index=False, header=False, sep='\t')

if __name__ == "__main__":
	
	main()






